'''Classes for implementing reactions with respect to some set of reactant RDKit Mols'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import ClassVar, Generator, Iterable, Optional, Type
from dataclasses import dataclass, field
from itertools import chain

from rdkit import Chem
from rdkit.Chem.rdchem import Mol

from .reactexc import ReactantTemplateMismatch
from .reactions import AnnotatedReaction, RxnProductInfo
from .fragment import IBIS, IntermonomerBondIdentificationStrategy, ReseparateRGroups, ReseparateRGroupsUnique

from .. import rdprops
from ..labeling import bondwise, molwise

from ...polymers.monomers.specification import SANITIZE_AS_KEKULE
from ...genutils.decorators.functional import optional_in_place


# REACTOR BASE CLASS 
@dataclass
class Reactor:
    '''Class for executing a reaction template on collections of RDKit Mol "reactants"'''
    rxn_schema : AnnotatedReaction
    _ridx_prop_name : ClassVar[str] = field(init=False, default='reactant_idx') # name of the property to assign reactant indices to; set for entire class

    # PRE-REACTION PREPARATION METHODS
    def _activate_reaction(self) -> None:
        '''Check that the reaction schema provided is well defined and initialized'''
        pass

    def __post_init__(self) -> None:
        '''Pre-processing of reaction and reactant Mols'''
        self._activate_reaction()

    @staticmethod
    @optional_in_place
    def _label_reactants(reactants : Iterable[Mol], reactant_label : str) -> None:
        '''Assigns "reactant_idx" Prop to all reactants to help track where atoms go during the reaction'''
        for i, reactant in enumerate(reactants):
            for atom in reactant.GetAtoms():
                atom.SetIntProp(reactant_label, i)

    # POST-REACTION CLEANUP METHODS
    @staticmethod
    @optional_in_place
    def _relabel_reacted_atoms(product : Mol, reactant_label : str, reactant_map_nums : dict[int, int]) -> None:
        '''Re-assigns "reactant_idx" Prop to modified reacted atoms to re-complete atom-to-reactant numbering'''
        for atom_id in rdprops.atom_ids_with_prop(product, 'old_mapno'):
            atom = product.GetAtomWithIdx(atom_id)
            map_num = atom.GetIntProp('old_mapno')

            atom.SetIntProp(reactant_label, reactant_map_nums[map_num])
            atom.SetAtomMapNum(map_num) # TOSELF : in future, might remove this (makes mapping significantly easier, but is ugly for labelling)

    @staticmethod
    @optional_in_place
    def _sanitize_bond_orders(product : Mol, product_template : Mol, product_info : RxnProductInfo) -> None:
        '''Ensure bond order changes specified by the reaction are honored by RDKit'''
        for prod_bond_id, map_num_pair in product_info.mod_bond_ids_to_map_nums.items():
            target_bond = product_template.GetBondWithIdx(prod_bond_id)

            product_bond = bondwise.get_bond_by_map_num_pair(product, map_num_pair)
            # product_bond = product.GetBondBetweenAtoms(*rdlabels.atom_ids_by_map_nums(product, *map_num_pair))
            assert(product_bond.GetBeginAtom().HasProp('_ReactionDegreeChanged')) 
            assert(product_bond.GetEndAtom().HasProp('_ReactionDegreeChanged')) # double check that the reaction agrees that the bond has changed

            product_bond.SetBondType(target_bond.GetBondType()) # set bond type to what it *should* be from the reaction schema

    # REACTION EXECUTION METHODS
    def react(self, reactants : Iterable[Mol], repetitions : int=1, clear_props : bool=False) -> list[Mol]:
        '''Execute reaction over a collection of reactants and generate product molecule(s)
        Does NOT require the reactants to match the order of the reacion template (only that some order fits)'''
        reactants = self.rxn_schema.valid_reactant_ordering(reactants, as_mols=True) # check that the reactants are compatible with the reaction
        if reactants is None:
            raise ReactantTemplateMismatch(f'Reactants provided to {self.__class__.__name__} are incompatible with reaction schema defined')
        reactants = self._label_reactants(reactants, reactant_label=self._ridx_prop_name, in_place=False) # assign reactant indices (not in-place)
        
        products : list[Mol] = []
        raw_products = self.rxn_schema.RunReactants(reactants, maxProducts=repetitions) # obtain unfiltered RDKit reaction output. TODO : generalize to work when more than 1 repetition is requested
        for i, product in enumerate(chain.from_iterable(raw_products)): # clean up products into a usable form
            self._relabel_reacted_atoms(
                product,
                reactant_label=self._ridx_prop_name,
                reactant_map_nums=self.rxn_schema.map_nums_to_reactant_nums,
                in_place=True
            )
            self._sanitize_bond_orders(product,
                product_template=self.rxn_schema.GetProductTemplate(i),
                product_info=self.rxn_schema.product_info_maps[i],
                in_place=True
            )
            if clear_props:
                rdprops.clear_atom_props(product, in_place=True)
            products.append(product)
        return products


# REACTOR SUBCLASSES
@dataclass
class AdditionReactor(Reactor):
    '''Special case of Reactor with two reactant species forming one product'''
    def __post_init__(self) -> None:
        assert(self.rxn_schema.GetNumReactantTemplates() == 2)
        assert(self.rxn_schema.GetNumProductTemplates() == 1)

        return super().__post_init__()
    
    @property
    def product_info(self) -> Mol:
        return self.rxn_schema.product_info_maps[0]

    def react(self, reactants : Iterable[Mol], repetitions : int = 1, clear_props : bool = False) -> Optional[Mol]:
        products = super().react(reactants, repetitions, clear_props) # return first (and only) product as standalone molecule
        if products:
            return products[0]

@dataclass
class CondensationReactor(Reactor):
    '''Special case of Reactor with two reactant species forming one product plus a small-molecule side product'''
    pass # TODO : implement behavior here

@dataclass
class PolymerizationReactor(Reactor):
    '''Reactor which exhaustively generates monomers fragments according to a given a polymerization mechanism'''
    def propagate(self, monomers : Iterable[Mol], fragment_strategy : IBIS=ReseparateRGroupsUnique(), clear_map_nums : bool=True) -> Generator[tuple[list[Mol], list[Mol]], None, None]:
        '''Keep reacting and fragmenting a pair of monomers until all reactive sites have been reacted
        Returns fragment pairs at each step of the chain propagation process'''
        reactants = monomers # initialize reactive pair with monomers
        while True: # check if the reactants can be applied under the reaction template
            try:
                intermediates = self.react(reactants, repetitions=1, clear_props=False) # can't clear properties yet, otherwise intermonomer bond finder would have nothing to go off of
            except ReactantTemplateMismatch:
                break
            
            fragments : list[Mol] = []
            for i, product in enumerate(intermediates):
                Chem.SanitizeMol(product, sanitizeOps=SANITIZE_AS_KEKULE) # clean up molecule, specifically avoiding de-kekulization in the case of aromatics
                if clear_map_nums:
                    molwise.clear_atom_map_nums(product, in_place=True)
                
                fragments.extend( # list extension preserves insertion order at each step
                    rdprops.clear_atom_props(fragment, in_place=False) # essential to avoid reaction mapping info from prior steps from contaminating future ones
                        for fragment in fragment_strategy.produce_fragments(
                            product,
                            product_info=self.rxn_schema.product_info_maps[i],
                            separate=True
                        )
                )
            yield intermediates, fragments # yield the dimerized fragment and the 2 new reactive fragments
            reactants = fragments # set fragments from current round of polymerizatio as reactants for next round


'''Unit tests for `molchargers` package'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import pytest

# NOTE: inverted custom imports here to get polymerist.openfftools import first, 
# Done so that if OpenFF is not found, a helpful installation error with be raised prior to attempting direct openff.toolkit imports below
from polymerist.genutils.importutils.dependencies import modules_installed
from polymerist.mdtools.openfftools.partialcharge.molchargers import MolCharger, ABE10Charger, EspalomaCharger, NAGLCharger
from polymerist.mdtools.openfftools.partialcharge.rescharge.interface import LibraryCharger

from openff.toolkit import Molecule
from openff.toolkit.utils.toolkits import OPENEYE_AVAILABLE


# Test MolCharger subclass registration
def test_molcharger_registers_subclasses() -> None:
    '''Test that the MolCharger tracks subclasses'''
    assert hasattr(MolCharger, 'subclass_registry')
    
@pytest.mark.parametrize('expected_charge_method_name, molcharger_subclass', MolCharger.subclass_registry.items()) # NOTE: this will fail if test_molcharger_registers_subclasses() fails
def test_molcharger_subclass_attr_registration(molcharger_subclass : type[MolCharger], expected_charge_method_name : str) -> None:
    '''Test that all MolCharger subclasses define and are registered under their "CHARGING_METHOD" class property'''
    assert hasattr(molcharger_subclass, 'CHARGING_METHOD') and (getattr(molcharger_subclass, 'CHARGING_METHOD') == expected_charge_method_name)


# Test MolCharger subclass implementations
@pytest.fixture
def offmol() -> Molecule:
    '''Dummy Molecule object for testing'''
    # DEV: worthing double-checking that partial charges are initially empty??
    return Molecule.from_smiles('c1ccccc1C(=O)O') # benzoic acid - nice and small, but with some non-trivial structure
    
## selectively register test to avoid failures due to missing optional dependencies
MOLCHARGER_TYPES_TO_TEST : list[type[MolCharger]] = [] 
if modules_installed('openeye.oechem', 'openeye.oeomega') and OPENEYE_AVAILABLE: # extra check needed block check when missing license (as is the case for the open-source polymerist repo)
    MOLCHARGER_TYPES_TO_TEST.append(ABE10Charger)
if modules_installed('espaloma_charge'):
    MOLCHARGER_TYPES_TO_TEST.append(EspalomaCharger)
if modules_installed('openff.nagl', 'openff.nagl_models'):
    MOLCHARGER_TYPES_TO_TEST.append(NAGLCharger)
# MOLCHARGER_TYPES_TO_TEST.append(
    # LibraryCharger  # LibraryCharger behave differently to other MolCharger and are kind of a pain in the ass generally...
# )                   # ...intend to deprecate and revamp them eventually, so will just exclude them from testing for now
    
@pytest.mark.parametrize('molcharger_subclass', MOLCHARGER_TYPES_TO_TEST)
def test_molchargers_assign_charges(offmol : Molecule, molcharger_subclass : type[MolCharger]) -> None:
    charger = molcharger_subclass()
    cmol = charger.charge_molecule(offmol)
    assert cmol.partial_charges is not None # should assign charges to the new, copied molecule

@pytest.mark.parametrize('molcharger_subclass', MOLCHARGER_TYPES_TO_TEST)
def test_molchargers_act_readonly(offmol : Molecule, molcharger_subclass : type[MolCharger]) -> None:
    charger = molcharger_subclass()
    cmol = charger.charge_molecule(offmol)
    assert offmol.partial_charges is None # should NOT affect the 

@pytest.mark.parametrize('molcharger_subclass', MOLCHARGER_TYPES_TO_TEST)
def test_molchargers_record_charge_method(offmol : Molecule, molcharger_subclass : type[MolCharger]) -> None:
    charger = molcharger_subclass()
    cmol = charger.charge_molecule(offmol)
    
    recorded_charge_method = cmol.properties.get('charge_method', None)
    assert (recorded_charge_method is not None) and (recorded_charge_method == getattr(molcharger_subclass, 'CHARGING_METHOD'))



'''For inserting text into other text in a rules-based manner'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import re


def insert_into_text_periodic(text : str, period : int, insertion : str='\n') -> str:
    '''Takes a string of text and another "insertion" string and inserts it throughout the text every <period> characters'''
    return insertion.join(text[i:i+period] for i in range(0, len(text), period))

def insert_into_text_periodic_re(text : str, period : int, insertion : str='\n') -> str:
    '''Takes a string of text and another "insertion" string and inserts it throughout the text every <period> characters
    Same as insert_into_text_periodic(), but implemented with regular expressions (allows for more complicated logical extensions)'''
    SPACE_RE = re.compile(f'(?s)(.{{{period}}})') # double curly braces escape the f-string syntax (to use as literals in regex quanitifer)
    return re.sub(SPACE_RE, f'\\1{insertion}', text)


'''Utilities for categorizing and representing file extensions/suffixes'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import mimetypes
from collections import defaultdict


# CUSTOM EXCEPTIONS
class FileTypeError(Exception):
    '''Raise when file extension is not valid for a particular application'''
    pass

# GLOBAL REGISTRY OF COMMON EXTENSIONS
EXT_REG = defaultdict(dict) # registry of builtin extensions for common file types
for ext, desc in mimetypes.types_map.items():
    ext_type, app = desc.split('/')
    EXT_REG[ext_type][ext] = app


'''Unit tests for `attrs` package'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from polymerist.genutils.attrs import compile_argfree_getable_attrs
import pytest


class ArgfreeGettableAttrTest():
    '''Dummy class for testing that dynamic attribute inspection works properly'''
    FOO = 'bar'

    def __init__(self, answer : int=42, spam : str='eggs') -> None:
        self.answer = answer
        self.spam = spam

    def get_answer(self) -> int:
        '''Prototypical getter which returns an attribute of self and is prefixed by "get"'''
        return self.answer
    
    def stringy_answer(self) -> str:
        '''A getter whose name does NOT contain any form of "get", but which nevertheless qualifies as argument-free'''
        return str(self.answer)
    
    def GetSpam(self) -> str:
        '''Camel-case getter to check that this is correctly handled by regex'''
        return self.spam

    @property
    def get_answer_prop(self) -> str:
        '''Property getter, to test that inspection excludes this'''
        return self.answer
    
    @classmethod
    def get_foo(cls) -> str:
        '''Class attr-based getter to test how shared-namespace attrs are handled'''
        return cls.FOO

    def _get_spam(self) -> int:
        '''Getter which is marked "private" and should not be returned '''
        return self.spam
    
    def get_echo(self, val : str) -> str:
        '''Getter which requires an argument passed and is therefore not argument-free'''
        return val

@pytest.fixture
def testobj() -> ArgfreeGettableAttrTest:
    return ArgfreeGettableAttrTest(answer=57, spam='ham')


# NOTE: not checking with getter_re='get' as this also returns __getstate__, which is somewhat unexpected to handle

def test_getable_attrs_regex_any(testobj):
    '''Test if getable arg regex match hits anywhere in method name'''
    attrs = compile_argfree_getable_attrs(testobj, getter_re='answer') # regex match should also hit in MIDDLE of string
    assert attrs == {'get_answer': testobj.answer, 'stringy_answer' : str(testobj.answer)}

def test_getable_attrs_regex_start(testobj):
    '''Test if getable arg regex match hits only at the start of in method name'''
    attrs = compile_argfree_getable_attrs(testobj, getter_re='^get') # regex match should also hit in MIDDLE of string
    assert attrs == {'get_answer': testobj.answer, 'get_foo' : testobj.FOO}

def test_getable_attrs_regex_case_insensitive(testobj):
    '''Test if getable arg regex match hits find method with regex only at the start of in method name irrespective of case'''
    attrs = compile_argfree_getable_attrs(testobj, getter_re='^[gG]et')#, getter_re='get')
    assert attrs == {'GetSpam': testobj.spam, 'get_answer': testobj.answer, 'get_foo': testobj.FOO}

def test_getable_attrs_regex_camel(testobj):
    '''Test if getable arg regex match hits camel-cased '''
    attrs = compile_argfree_getable_attrs(testobj, getter_re='^Get[A-Z].*') # regex match should also hit in MIDDLE of string
    assert attrs == {'GetSpam' : testobj.spam}

def test_getable_attrs_regex_repl(testobj):
    '''Test if regex replacement on returned method names is performed as expected'''
    repl = 'TEST'
    attrs = compile_argfree_getable_attrs(testobj, getter_re='^get', repl_str=repl) # regex match should also hit in MIDDLE of string
    assert attrs == {f'{repl}_answer': testobj.answer, f'{repl}_foo': testobj.FOO}